import os
import logging
import json
from datetime import datetime, timedelta
import pandas as pd
from dotenv import load_dotenv
import time
import requests
import re

# 로깅 설정
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# .env 파일에서 환경 변수 로드
CONFIG_DIR = 'C:\\Users\\user\\projects\\st\\config'
load_dotenv(dotenv_path=os.path.join(CONFIG_DIR, '.env'))

# 상수 설정
RETRIES = 3
DELAY = 60  # 재시도 사이의 지연 시간(초)
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
SHORT_DATE_FORMAT = "%Y%m%d"
MAX_ATTEMPTS = 10
PAGINATION_DELAY = 0.05  # 페이지네이션 API 요청 사이의 지연 시간(초)

TRADING_START_HOUR = 9
TRADING_END_HOUR = 15  # Typically, trading hours might be 9 AM to 3 PM, adjust as necessary.

# 설정 로드 함수
def load_settings():
    settings = {
        'data_path': os.getenv('DATA_PATH'),
        'model_path': os.getenv('MODEL_PATH'),
        'approval_key': os.getenv('APPROVAL_KEY'),
        'max_detected_stocks': int(os.getenv('MAX_DETECTED_STOCKS', 3000)),  # 기본값을 높게 설정
        'data_fetch_interval': int(os.getenv('DATA_FETCH_INTERVAL', 60)),
        'strategy_name': os.getenv('STRATEGY_NAME'),
        'token_path': os.getenv('TOKEN_PATH'),
        'appkey': os.getenv('APPKEY'),
        'appsecret': os.getenv('APPSECRET'),
        'user_id': os.getenv('USER_ID'),
        'trading_url': os.getenv('TRADING_URL'),
        'agent': os.getenv('AGENT'),
        'acct_stock': os.getenv('ACCT_STOCK'),
        'prod': os.getenv('PROD'),
        'cano': os.getenv('CANO')
    }

    config_json_path = os.path.join(CONFIG_DIR, 'config.json')
    if (os.path.exists(config_json_path)):
        with open(config_json_path, 'r') as f:
            config_settings = json.load(f)
            settings.update(config_settings)
    else:
        raise FileNotFoundError(f"{config_json_path} 파일을 찾을 수 없습니다.")
    
    return settings

settings = load_settings()

# 거래 시간 확인 함수
def is_trading_hours():
    """현재 시간이 거래 시간인지 확인합니다."""
    now = datetime.now()
    return TRADING_START_HOUR <= now.hour <= TRADING_END_HOUR

# 토큰 처리 함수
def get_token_data():
    """토큰 파일에서 토큰 데이터를 가져옵니다."""
    token_path = settings['token_path']
    try:
        with open(token_path, 'r') as file:
            return json.load(file)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        logging.error(f"토큰 파일 읽기 오류: {e}")
        return None

def save_token_data(token_data):
    """토큰 데이터를 토큰 파일에 저장합니다."""
    token_path = settings['token_path']
    os.makedirs(os.path.dirname(token_path), exist_ok=True)
    with open(token_path, 'w') as file:
        json.dump(token_data, file)

def is_token_valid(token_data):
    """토큰의 만료 시간을 기준으로 유효성을 검사합니다."""
    if not token_data or 'access_token' not in token_data:
        return False
    expiration_str = token_data.get('expiration')
    if not expiration_str:
        return False
    expiration = datetime.strptime(expiration_str, DATE_FORMAT)
    return expiration > datetime.now()

def get_valid_token():
    """유효한 토큰을 가져오고, 필요시 갱신합니다."""
    token_data = get_token_data()
    if is_token_valid(token_data):
        logging.info("기존 토큰 사용 중")
        return token_data['access_token']
    else:
        logging.info("토큰이 만료되었거나 유효하지 않습니다. 갱신 중...")
        return refresh_token()

def refresh_token(retries=RETRIES, delay=DELAY):
    """액세스 토큰을 갱신하며, 실패 시 재시도합니다."""
    url = f"{settings['trading_url']}/oauth2/tokenP"
    data = {
        'grant_type': 'client_credentials',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret']
    }
    headers = {'Content-Type': 'application/json'}
    
    for attempt in range(retries):
        try:
            logging.info("새 토큰 요청 중...")
            response = requests.post(url, headers=headers, json=data)
            if response.status_code == 200:
                response_data = response.json()
                new_token = response_data.get('access_token')

                # expires_in이 '2024-08-26 18:14:17'와 같은 날짜 문자열일 수 있음
                expires_in = response_data.get('access_token_token_expired')
                
                try:
                    expiration = datetime.strptime(expires_in, DATE_FORMAT)  # 날짜 형식으로 변환 시도
                except ValueError:
                    expiration = datetime.now() + timedelta(seconds=int(expires_in) if expires_in else 3600)
                
                token_data = {'access_token': new_token, 'expiration': expiration.strftime(DATE_FORMAT)}
                save_token_data(token_data)
                logging.info(f"새 토큰 저장됨. 만료 시간: {expiration}")
                return new_token
            else:
                logging.error(f"토큰 갱신 오류: {response.text}")
                if "EGW00133" in response.text:
                    logging.info(f"{delay}초 후 재시도...")
                    time.sleep(delay)
                else:
                    return None
        except requests.exceptions.RequestException as e:
            logging.error(f"토큰 갱신 중 요청 예외 발생: {e}")
            return None
    return None

# API 요청 함수
def get_api_response(url, headers, params, max_retries=3):
    for attempt in range(max_retries):
        try:
            response = requests.get(url, headers=headers, params=params)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            logging.error(f"요청 예외 발생 (시도 {attempt + 1}/{max_retries}): {e}")
            time.sleep(2 ** attempt)  # 지수적으로 증가하는 지연 시간을 두고 재시도
    return None


def fetch_stock_data(stock_code, start_date, end_date):
    """주어진 종목 코드와 날짜 범위에 대한 주식 데이터를 가져옵니다."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/inquire-daily-itemchartprice"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret'],
        'tr_id': 'FHKST03010100'
    }
    params = {
        'FID_COND_MRKT_DIV_CODE': 'J',
        'FID_INPUT_ISCD': stock_code,
        'FID_INPUT_DATE_1': start_date.strftime(SHORT_DATE_FORMAT),
        'FID_INPUT_DATE_2': end_date.strftime(SHORT_DATE_FORMAT),
        'FID_PERIOD_DIV_CODE': 'D',
        'FID_ORG_ADJ_PRC': '0'
    }

    return get_api_response(url, headers, params)

# 검색 함수
def get_search_conditions():
    """API로부터 검색 조건을 가져옵니다."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/psearch-title"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret'],
        'tr_id': 'HHKST03900300'
    }
    params = {'user_id': settings['user_id']}
    logging.info(f"요청 URL: {url}")
    logging.info(f"요청 헤더: {headers}")
    logging.info(f"요청 파라미터: {params}")

    search_conditions = []
    attempt_count = 0
    next_page_token = None

    while attempt_count < MAX_ATTEMPTS:
        if next_page_token:
            params['next_key'] = next_page_token

        response = get_api_response(url, headers, params)
        logging.info(f"API 응답: {response}")

        if response and 'output2' in response:
            search_conditions.extend(response['output2'])
            break
        elif response and response.get('msg_cd') == 'MCA05762':
            logging.info("페이지네이션 처리 중.")
            attempt_count += 1
            next_page_token = response.get('next_key')
            if not next_page_token:
                logging.error("다음 키가 예상되었지만 응답에 없음.")
                break
        else:
            logging.error(f"예상치 못한 응답: {response}")
            break

    if attempt_count >= MAX_ATTEMPTS:
        logging.error("최대 페이지네이션 시도 도달. 종료.")
    logging.info(f"검색 조건: {search_conditions}")
    return search_conditions if search_conditions else None

def get_search_results(seq):
    """제공된 시퀀스를 사용하여 API로부터 검색 결과를 가져옵니다."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/psearch-result"
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'Bearer {get_valid_token()}',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret'],
        'tr_id': 'HHKST03900400'
    }
    params = {'user_id': settings['user_id'], 'seq': seq}

    logging.info(f"요청 URL: {url}")
    logging.info(f"요청 헤더: {headers}")
    logging.info(f"요청 파라미터: {params}")

    response = get_api_response(url, headers, params)
    logging.info(f"API 응답: {response}")

    if isinstance(response, dict):
        if 'output2' in response:
            return response['output2']
        else:
            if response.get('msg_cd') == 'MCA05918':
                logging.error(f"조건 검색 결과 0개 종목. 전체 응답: {response}")
            else:
                logging.error(f"응답에 'output2' 없음: {response}")
                logging.error(f"전체 응답: {response}")
            return None
    else:
        logging.error(f"예상치 못한 응답 형식: {response}")
        return None

# 데이터 처리 함수
def load_and_filter_stock_data(file_path):
    try:
        df = pd.read_csv(file_path)
        df = df[df['code'].apply(lambda x: bool(re.match(r'^\d+$', str(x))))]
        if df.empty:
            logging.error(f"All stock codes in {file_path} were filtered out or the file is empty.")
            return None
        logging.info(f"Filtered stock data:\n{df.head()}")
        return df
    except FileNotFoundError:
        logging.error(f"File not found: {file_path}")
        return None
    except pd.errors.EmptyDataError:
        logging.error(f"The file is empty: {file_path}")
        return None
    except pd.errors.ParserError:
        logging.error(f"Parsing error occurred while reading {file_path}")
        return None
    except Exception as e:
        logging.error(f"An error occurred while processing {file_path}: {e}")
        return None


def save_data_to_csv(data, stock_code):
    if not re.match(r'^\d+$', stock_code):
        logging.error(f"Skipping stock code {stock_code} because it contains non-numeric characters.")
        return

    file_path = os.path.join(settings['data_path'], 'stock_data', f"{stock_code.zfill(6)}_data.csv")
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    if 'output2' in data:
        output_data = []
        for item in data['output2']:
            try:
                date_value = item.get('stck_bsop_date')
                if date_value:
                    # 날짜가 None이 아닌 경우에만 변환을 시도
                    date = pd.to_datetime(date_value, format='%Y%m%d', errors='coerce')
                    if pd.isna(date):
                        logging.error(f"Invalid date '{date_value}' for stock code {stock_code}. Skipping this record.")
                        continue
                    date = date.strftime('%Y-%m-%d')
                else:
                    logging.error(f"Missing date for stock code {stock_code}. Skipping this record.")
                    continue  # 날짜가 없으면 해당 레코드를 건너뜁니다.

                output2_data = {
                    'open': int(item.get('stck_oprc', '0').replace(',', '')) if item.get('stck_oprc') else 0,
                    'high': int(item.get('stck_hgpr', '0').replace(',', '')) if item.get('stck_hgpr') else 0,
                    'low': int(item.get('stck_lwpr', '0').replace(',', '')) if item.get('stck_lwpr') else 0,
                    'close': int(item.get('stck_clpr', '0').replace(',', '')) if item.get('stck_clpr') else 0,
                    'volume': int(item.get('acml_tr_pbmn', '0').replace(',', '')) if item.get('acml_tr_pbmn') else 0,
                    'stck_bsop_date': date,  
                    'corp_code': stock_code
                }
                output_data.append(output2_data)
            except Exception as e:
                logging.error(f"Error processing record for stock code {stock_code}: {e}")
                continue

        if output_data:
            new_df = pd.DataFrame(output_data)
            if os.path.exists(file_path):
                new_df.to_csv(file_path, mode='a', header=False, index=False)
            else:
                new_df.to_csv(file_path, mode='w', index=False)
            logging.info(f"주식 데이터가 {file_path}에 저장되었습니다.")
        else:
            logging.info(f"No new valid data to save for stock code {stock_code}.")
    else:
        logging.error("CSV로 저장하기 위한 데이터 형식이 인식되지 않았습니다.")

def get_stock_codes_and_names(seq):
    """검색 결과에서 종목 코드와 이름을 가져옵니다."""
    search_conditions = get_search_conditions()
    if not search_conditions:
        logging.error("검색 조건을 찾을 수 없습니다.")
        return []

    search_results = get_search_results(seq)
    if not search_results:
        logging.error("검색 결과를 찾을 수 없습니다.")
        return []

    stock_codes_and_names = []
    for result in search_results:
        code = result.get('code')
        name = result.get('name')
        logging.info(f"가져온 코드: {code}, 이름: {name}")
        if code and name:
            stock_codes_and_names.append((code.zfill(6), name))
        else:
            logging.error(f"유효하지 않은 종목 코드 또는 이름: {result}")

    logging.info(f"검색된 종목 코드와 이름: {stock_codes_and_names}")
    
    # 검색된 종목 코드와 이름을 stock_n_codes.csv에 저장
    save_stock_codes_to_csv(stock_codes_and_names, 'stock_n_codes.csv')
    
    # 모든 검색된 종목 코드를 stock_codes.csv에 누적 저장
    save_stock_codes_to_csv([code for code, name in stock_codes_and_names], 'stock_codes.csv', append=True)
    
    return stock_codes_and_names

def save_stock_codes_to_csv(stock_codes_and_names, filename, append=False):
    """종목 코드와 이름을 주어진 파일에 저장합니다."""
    file_path = os.path.join(settings['data_path'], filename)
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    
    logging.info(f"저장할 데이터 개수: {len(stock_codes_and_names)}")
    logging.info(f"파일 경로: {file_path}")
    
    if len(stock_codes_and_names) == 0:
        logging.warning("저장할 데이터가 없습니다. 파일이 생성되지 않습니다.")
        return
    
    if filename == 'stock_codes.csv':
        new_df = pd.DataFrame(stock_codes_and_names, columns=['code'])
    else:
        new_df = pd.DataFrame(stock_codes_and_names, columns=['code', 'name'])
    
    if os.path.exists(file_path) and append:
        try:
            existing_df = pd.read_csv(file_path)
        except Exception as e:
            logging.error(f"기존 파일을 읽는 중 오류 발생: {e}")
            if filename == 'stock_codes.csv':
                existing_df = pd.DataFrame(columns=['code'])
            else:
                existing_df = pd.DataFrame(columns=['code', 'name'])
        
        combined_df = pd.concat([existing_df, new_df]).drop_duplicates().reset_index(drop=True)
    else:
        combined_df = new_df
    
    combined_df.to_csv(file_path, mode='w', index=False)
    logging.info(f"종목 코드가 {file_path}에 저장되었습니다.")
# 데이터 가져오기 및 저장 함수
def fetch_and_save_data(stock_codes_and_names):
    """검색된 종목 코드와 이름에 대한 데이터를 가져오고 저장합니다."""
    token_data = get_token_data()
    access_token = token_data['access_token'] if token_data else None

    today = datetime.now().date()
    ten_years_ago = today - timedelta(days=365 * 10)
    date_range = pd.date_range(end=today, start=ten_years_ago, freq='MS')

    batch_size = 100  # 한 번에 처리할 종목 수
    num_batches = len(stock_codes_and_names) // batch_size + (1 if len(stock_codes_and_names) % batch_size != 0 else 0)

    for batch_idx in range(num_batches):
        batch_start = batch_idx * batch_size
        batch_end = batch_start + batch_size
        batch = stock_codes_and_names[batch_start:batch_end]

        for start_date, end_date in zip(date_range, date_range[1:]):
            logging.info(f"날짜 범위: {start_date}부터 {end_date}까지 데이터 가져오는 중")
            for stock_code, stock_name in batch:
                data = fetch_stock_data(stock_code, start_date, end_date)
                if data:
                    save_data_to_csv(data, stock_code)
                else:
                    logging.error(f"종목: {stock_code}에 대한 데이터를 가져오지 못했습니다.")

                time.sleep(PAGINATION_DELAY)  # API 요청 사이의 지연 시간

def fetch_data_for_detected_stocks():
    """검색된 종목에 대한 데이터를 가져오는 메인 함수."""
    if is_trading_hours():
        logging.info("거래 시간입니다. 검색된 종목에 대한 데이터를 가져오는 중...")
        for seq in range(2):
            detected_stocks = get_stock_codes_and_names(seq)
            if detected_stocks:
                fetch_and_save_data(detected_stocks)
                break
            else:
                logging.info(f"시퀀스 {seq}에 대해 찾은 종목이 없습니다. 다음 시퀀스를 시도합니다.")
    else:
        logging.info("거래 시간이 아닙니다. stock_codes.csv에서 종목 코드를 로드하여 데이터를 가져옵니다.")
        stock_codes = load_stock_codes_from_csv()
        if not stock_codes:
            logging.error("stock_codes.csv에서 로드된 유효한 종목 코드가 없습니다. 실행을 종료합니다.")
            return
        stock_codes_and_names = [(code, '') for code in stock_codes]  # 이름이 필요하지 않으므로 빈 문자열로 채움
        fetch_and_save_data(stock_codes_and_names)

def load_stock_codes_from_csv():
    """stock_codes.csv 파일에서 종목 코드를 로드합니다."""
    stock_data_file_path = os.path.join(settings['data_path'], 'stock_codes.csv')
    filtered_stock_data = load_and_filter_stock_data(stock_data_file_path)
    if filtered_stock_data is not None:
        return filtered_stock_data['code'].tolist()
    else:
        return []

if __name__ == "__main__":
    fetch_data_for_detected_stocks()
