import os
import pandas as pd
from sklearn.preprocessing import StandardScaler
import logging
from sklearn.model_selection import train_test_split
import sys

# 프로젝트 루트 및 스크립트 경로 설정
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
scripts_dir = os.path.join(project_root, 'scripts')
config_dir = os.path.join(project_root, 'config')
utils_dir = os.path.join(scripts_dir, 'utils')

sys.path.extend([
    scripts_dir,
    os.path.join(scripts_dir, 'data_collection'),
    os.path.join(scripts_dir, 'data_processing'),
    os.path.join(scripts_dir, 'model'),
    os.path.join(scripts_dir, 'trading'),
    utils_dir
])

logging.basicConfig(level=logging.INFO)

def normalize_data(df):
    """Normalize the data using StandardScaler."""
    scaler = StandardScaler()
    normalized_df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns)
    return normalized_df, scaler

def load_csv_files_from_directory(directory_path, identifier_col=None):
    """디렉토리 내 CSV 파일을 로드하는 함수"""
    dataframes = []
    for file_name in os.listdir(directory_path):
        if file_name.endswith('.csv'):
            file_path = os.path.join(directory_path, file_name)
            try:
                df = pd.read_csv(file_path, encoding='utf-8')
                logging.info(f"Loaded {file_name} with columns: {df.columns}")
                df.columns = df.columns.str.lower()

                # 표준화된 컬럼 이름으로 변경
                if 'date' in df.columns:
                    df.rename(columns={'date': 'stck_bsop_date'}, inplace=True)

                if identifier_col and identifier_col not in df.columns:
                    ticker = os.path.splitext(file_name)[0].split('_')[0]
                    df[identifier_col] = ticker

                if not df.empty:
                    logging.info(f"First few rows of {file_name}:\n{df.head()}")
                    dataframes.append(df)
                else:
                    logging.warning(f"Empty file: {file_path}")
            except pd.errors.EmptyDataError:
                logging.warning(f"Empty file: {file_path}")
            except Exception as e:
                logging.error(f"Error reading {file_path}: {e}")
    if dataframes:
        return pd.concat(dataframes, ignore_index=True)
    else:
        logging.error(f"No valid CSV files found in directory: {directory_path}")
        return pd.DataFrame()

def preprocess_df(df, date_column='stck_bsop_date', date_format='%Y-%m-%d'):
    """데이터프레임 전처리 함수: 날짜 변환 및 데이터 정리"""
    if date_column not in df.columns:
        raise ValueError(f"Date column '{date_column}' not found in the dataset. Available columns: {df.columns}")

    logging.info(f"Before date conversion, first few rows:\n{df[[date_column]].head()}")

    df = df.copy()

    # Convert the date column to string and strip any whitespace or unexpected characters
    df[date_column] = df[date_column].astype(str).str.strip()

    # Identify and log any anomalies in date values before conversion
    unique_dates = df[date_column].unique()
    logging.info(f"Unique date formats before conversion:\n{unique_dates}")

    # Replace any non-standard characters, e.g., replace non-breaking spaces with normal spaces
    df[date_column] = df[date_column].str.replace('\xa0', ' ', regex=False)

    # Try to clean any additional unexpected characters
    df[date_column] = df[date_column].str.replace('[^\w\s-]', '', regex=True)

    # Attempt to convert dates, log any conversion errors
    try:
        df[date_column] = pd.to_datetime(df[date_column], format=date_format, errors='coerce')
    except Exception as e:
        logging.error(f"Date conversion failed: {e}")

    # Log rows where date conversion failed
    failed_conversion = df[df[date_column].isna()]
    if not failed_conversion.empty:
        logging.error(f"Some dates could not be converted for column {date_column}. Rows:\n{failed_conversion[[date_column]].head()}")
        # Attempt a flexible date parsing
        df[date_column] = pd.to_datetime(df[date_column], errors='coerce')

    # Drop rows with invalid dates
    df.dropna(subset=[date_column], inplace=True)
    
    if df.empty:
        logging.warning(f"Data is empty after dropping rows with invalid dates.")
    
    df.fillna(0, inplace=True)
    return df

def preprocess_data(stock_data, financial_data=None, economic_data=None, gdp_data=None, inflation_data=None, news_data=None):
    """데이터 전처리 및 병합 함수"""
    if stock_data.empty:
        logging.error("Stock data is empty. Cannot proceed with preprocessing.")
        return None
    
    logging.info(f"Initial stock data shape: {stock_data.shape}")
    
    # Check for the correct date column in stock_data
    date_column = 'stck_bsop_date' if 'stck_bsop_date' in stock_data.columns else 'date'
    
    stock_data = preprocess_df(stock_data, date_column=date_column, date_format='%Y%m%d')
    
    if stock_data.empty:
        logging.error("Preprocessed stock data is empty after date conversion.")
        return None
    
    logging.info(f"Stock data after preprocessing shape: {stock_data.shape}")
    
    # Preprocess financial data
    if financial_data is not None and not financial_data.empty:
        financial_data = preprocess_df(financial_data, date_column='stck_bsop_date', date_format='%Y%m%d')
        if financial_data.empty:
            logging.warning("Financial data is empty after preprocessing.")
            financial_data = None
    
    # Preprocess economic data
    if economic_data is not None and not economic_data.empty:
        economic_data = preprocess_df(economic_data, date_column='stck_bsop_date', date_format='%Y%m%d')
        if economic_data.empty:
            logging.warning("Economic data is empty after preprocessing.")
            economic_data = None

    # Preprocess GDP data
    if gdp_data is not None and not gdp_data.empty:
        gdp_data = preprocess_df(gdp_data, date_column='stck_bsop_date', date_format='%Y-%m-%d')
        if gdp_data.empty:
            logging.warning("GDP data is empty after preprocessing.")
            gdp_data = None

    # Preprocess inflation data
    if inflation_data is not None and not inflation_data.empty:
        inflation_data = preprocess_df(inflation_data, date_column='stck_bsop_date', date_format='%Y-%m-%d')
        if inflation_data.empty:
            logging.warning("Inflation data is empty after preprocessing.")
            inflation_data = None

    # Preprocess news data
    if news_data is not None and not news_data.empty:
        news_data = preprocess_df(news_data, date_column='stck_bsop_date', date_format='%Y-%m-%d')
        if news_data.empty:
            logging.warning("News data is empty after preprocessing.")
            news_data = None

    # Assuming 'close' is the closing price used to create the target
    stock_data['target'] = stock_data['close'].shift(-1)
    stock_data.dropna(subset=['target'], inplace=True)

    if stock_data.empty:
        logging.error("No valid stock data after creating target.")
        return None

    # Merging datasets
    try:
        if financial_data is not None and not financial_data.empty:
            stock_data = pd.merge_asof(stock_data.sort_values('stck_bsop_date'), financial_data.sort_values('stck_bsop_date'), on='stck_bsop_date', direction='nearest')
            logging.info(f"Stock data shape after merging with financial data: {stock_data.shape}")
        
        if economic_data is not None and not economic_data.empty:
            stock_data = pd.merge_asof(stock_data.sort_values('stck_bsop_date'), economic_data.sort_values('stck_bsop_date'), on='stck_bsop_date', direction='nearest')
            logging.info(f"Stock data shape after merging with economic data: {stock_data.shape}")
        
        if gdp_data is not None and not gdp_data.empty:
            stock_data = pd.merge_asof(stock_data.sort_values('stck_bsop_date'), gdp_data.sort_values('stck_bsop_date'), on='stck_bsop_date', direction='nearest')
            logging.info(f"Stock data shape after merging with GDP data: {stock_data.shape}")
        
        if inflation_data is not None and not inflation_data.empty:
            stock_data = pd.merge_asof(stock_data.sort_values('stck_bsop_date'), inflation_data.sort_values('stck_bsop_date'), on='stck_bsop_date', direction='nearest')
            logging.info(f"Stock data shape after merging with inflation data: {stock_data.shape}")
        
        if news_data is not None and not news_data.empty:
            stock_data = pd.merge_asof(stock_data.sort_values('stck_bsop_date'), news_data.sort_values('stck_bsop_date'), on='stck_bsop_date', direction='nearest')
            logging.info(f"Stock data shape after merging with news data: {stock_data.shape}")

    except Exception as e:
        logging.error(f"Failed to merge data for stock data: {e}")
        return None

    if stock_data.empty:
        logging.error("No valid combined data after merging.")
        return None

    for col in stock_data.columns:
        if stock_data[col].dtype == 'object':
            stock_data[col] = pd.to_numeric(stock_data[col].str.replace(',', ''), errors='coerce')

    stock_data.ffill(inplace=True)
    stock_data.bfill(inplace=True)
    stock_data.fillna(0, inplace=True)

    X = stock_data.drop(['target', 'stck_bsop_date'], axis=1)  # Drop the updated date column
    y = stock_data['target']

    if X.empty or y.empty:
        logging.error("Insufficient data after preprocessing for training the model.")
        return None

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)

    return (X_train, X_test, y_train, y_test), scaler, stock_data


def preprocess_and_save(stock_data, financial_data, economic_data, gdp_data, inflation_data, news_data, output_dir, corp_code):
    """전처리된 데이터를 저장하는 함수"""
    result = preprocess_data(stock_data, financial_data, economic_data, gdp_data, inflation_data, news_data)
    
    if result is None:
        logging.error(f"Data preprocessing failed for {corp_code}.")
        return

    (_, _, _, _), _, combined_data = result

    if combined_data is None or combined_data.empty:
        logging.error(f"No valid combined data after preprocessing for {corp_code}.")
        return

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    output_path = os.path.join(output_dir, f'preprocessed_data_{corp_code}.csv')
    combined_data.to_csv(output_path, index=False)
    logging.info(f"Preprocessed data for {corp_code} saved to {output_path}")

def main(base_dir):
    """메인 실행 함수"""
    stock_data_dir = os.path.join(base_dir, 'stock_data')
    financial_data_dir = os.path.join(base_dir, 'financial_data')
    economic_data_dir = os.path.join(base_dir, 'economic_data')
    gdp_data_dir = os.path.join(base_dir, 'gdp_data')
    inflation_data_dir = os.path.join(base_dir, 'inflation_data')
    news_data_dir = os.path.join(base_dir, 'news_data')

    stock_data = load_csv_files_from_directory(stock_data_dir, identifier_col='corp_code')
    if stock_data.empty:
        logging.error("Stock data is missing or empty after loading.")
        return
    logging.info(f"Stock data initial state: {stock_data.head()}")

    financial_data = load_csv_files_from_directory(financial_data_dir)
    economic_data = load_csv_files_from_directory(economic_data_dir)
    gdp_data = load_csv_files_from_directory(gdp_data_dir)
    inflation_data = load_csv_files_from_directory(inflation_data_dir)
    news_data = load_csv_files_from_directory(news_data_dir)

    if financial_data.empty:
        logging.warning("Financial data is missing or empty.")
    if economic_data.empty:
        logging.warning("Economic data is missing or empty.")
    if gdp_data.empty:
        logging.warning("GDP data is missing or empty.")
    if inflation_data.empty:
        logging.warning("Inflation data is missing or empty.")
    if news_data.empty:
        logging.warning("News data is missing or empty.")
    
    identifier_column = 'corp_code'
    if identifier_column in stock_data.columns:
        corp_codes = stock_data[identifier_column].unique()
    else:
        logging.error(f"'{identifier_column}' column not found. Cannot proceed without proper identification information.")
        return

    output_dir = os.path.join(base_dir, 'preprocessed_data')

    for corp_code in corp_codes:
        corp_stock_data = stock_data[stock_data[identifier_column] == corp_code]
        corp_financial_data = financial_data[financial_data[identifier_column] == corp_code] if not financial_data.empty else pd.DataFrame()

        preprocess_and_save(corp_stock_data, corp_financial_data, economic_data, gdp_data, inflation_data, news_data, output_dir, corp_code)

if __name__ == "__main__":
    base_dir = 'C:\\Users\\user\\projects\\st\\data'  # Use your data directory path
    main(base_dir)
