import os
import pandas as pd
import numpy as np
import logging
from model_training import train_lstm_model, update_lstm_model, build_lstm_model

# 데이터 경로 설정
DATA_DIR = "C:/Users/user/projects/st/data"
PREPROCESSED_DATA_DIR = os.path.join(DATA_DIR, "preprocessed_data")
MODEL_DIR = os.path.join(DATA_DIR, "models")
SEQUENCE_LENGTH = 60

# 로깅 설정
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

class ModelTrainer:
    def __init__(self):
        pass

    def train(self, X_train, y_train, X_val, y_val):
        input_shape = (X_train.shape[1], X_train.shape[2])
        model, history = train_lstm_model(X_train, y_train, X_val, y_val, input_shape)
        return model

    def update(self, model, X_train, y_train, X_val, y_val):
        model, history = update_lstm_model(model, X_train, y_train, X_val, y_val)
        return model

def load_data(file_path):
    df = pd.read_csv(file_path)
    df['date'] = pd.to_datetime(df['date'])
    
    feature_columns = [col for col in df.columns if col not in ['date', 'target']]
    df = df.sort_values('date')
    
    return df, feature_columns

def preprocess_data(df, feature_columns):
    df.fillna(method='ffill', inplace=True)
    df = df[df['volume'] > 0]

    train_data = df[:int(0.7 * len(df))]
    val_data = df[int(0.7 * len(df)):int(0.85 * len(df))]
    test_data = df[int(0.85 * len(df)):]

    return train_data, val_data, test_data, feature_columns

def create_sequences(data, seq_length, feature_columns):
    X, y = [], []
    for i in range(len(data) - seq_length):
        X.append(data.iloc[i:i+seq_length][feature_columns].values)
        y.append(data.iloc[i+seq_length]['target'])
    return np.array(X), np.array(y)

def main():
    for file_name in os.listdir(PREPROCESSED_DATA_DIR):
        if file_name.endswith('.csv'):
            corp_code = os.path.splitext(file_name)[0]
            file_path = os.path.join(PREPROCESSED_DATA_DIR, file_name)
            logging.info(f"Processing {corp_code}...")

            df, feature_columns = load_data(file_path)
            train_data, val_data, test_data, feature_columns = preprocess_data(df, feature_columns)
            
            X_train, y_train = create_sequences(train_data, SEQUENCE_LENGTH, feature_columns)
            X_val, y_val = create_sequences(val_data, SEQUENCE_LENGTH, feature_columns)
            X_test, y_test = create_sequences(test_data, SEQUENCE_LENGTH, feature_columns)
            
            trainer = ModelTrainer()

            lstm_model_path = os.path.join(MODEL_DIR, f"lstm_{corp_code}.h5")
            if os.path.exists(lstm_model_path):
                model = build_lstm_model((SEQUENCE_LENGTH, len(feature_columns)))
                model.load_weights(lstm_model_path)
                model = trainer.update(model, X_train, y_train, X_val, y_val)
            else:
                model = trainer.train(X_train, y_train, X_val, y_val)
            model.save(lstm_model_path)

            logging.info(f"Model saved for {corp_code}.")

if __name__ == "__main__":
    main()
