import os
import pandas as pd
import numpy as np
import tensorflow as tf
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import LSTM, Dense, Dropout
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
import logging
import joblib

# 로그 설정
logging.basicConfig(level=logging.INFO)

# 데이터 경로 설정
DATA_DIR = "C:/Users/user/projects/st/data"
PREPROCESSED_DATA_DIR = os.path.join(DATA_DIR, "preprocessed_data")
MODEL_DIR = os.path.join(DATA_DIR, "models")

SEQUENCE_LENGTH = 60

def load_and_preprocess_data(file_path):
    df = pd.read_csv(file_path)
    df['date'] = pd.to_datetime(df['date'])

    # 거래 정지된 날짜 확인 및 처리
    df['volume'].replace(0, np.nan, inplace=True)
    df['close'].replace(0, np.nan, inplace=True)
    df.fillna(method='ffill', inplace=True)  # 이전 값으로 결측값 채우기

    # 비정상적 데이터 필터링: 최소한의 거래량 조건 추가
    df = df[df['volume'] > 0]

    # 기타 전처리
    scaler = StandardScaler()
    feature_columns = [col for col in df.columns if col not in ['date']]
    df[feature_columns] = scaler.fit_transform(df[feature_columns])
    
    # 데이터 나누기
    train_data = df[:int(0.7 * len(df))]
    val_data = df[int(0.7 * len(df)):int(0.85 * len(df))]
    test_data = df[int(0.85 * len(df)):]

    return train_data, val_data, test_data, feature_columns, scaler

def create_sequences(data, seq_length, feature_columns):
    X, y = [], []
    for i in range(len(data) - seq_length):
        X.append(data.iloc[i:i+seq_length][feature_columns].values)
        y.append(data.iloc[i+seq_length]['target'])
    return np.array(X), np.array(y)

def build_and_train_model(X_train, y_train, X_val, y_val):
    model = Sequential()
    model.add(LSTM(64, input_shape=(X_train.shape[1], X_train.shape[2]), return_sequences=True))
    model.add(Dropout(0.2))
    model.add(LSTM(32, return_sequences=False))
    model.add(Dropout(0.2))
    model.add(Dense(1, activation='linear'))

    model.compile(optimizer='adam', loss='mse')

    early_stop = EarlyStopping(monitor='val_loss', patience=10, restore_best_weights=True)
    model_checkpoint = ModelCheckpoint('best_model.h5', monitor='val_loss', save_best_only=True)

    history = model.fit(
        X_train, y_train,
        epochs=100,
        batch_size=32,
        validation_data=(X_val, y_val),
        callbacks=[early_stop, model_checkpoint]
    )

    return model, history

def evaluate_model(model, X_test, y_test):
    y_pred = model.predict(X_test)

    mse = mean_squared_error(y_test, y_pred)
    mae = mean_absolute_error(y_test, y_pred)
    r2 = r2_score(y_test, y_pred)

    print(f"Test MSE: {mse:.4f}")
    print(f"Test MAE: {mae:.4f}")
    print(f"Test R2 Score: {r2:.4f}")

    return y_pred

def calculate_and_compare_returns(test_data, y_pred):
    actual_prices = test_data.iloc[SEQUENCE_LENGTH:]['close'].values
    signal = np.where(y_pred[:-1] < y_pred[1:], 1, -1)
    returns = np.diff(actual_prices) / actual_prices[:-1]
    strategy_returns = returns * signal
    cumulative_returns = np.cumprod(1 + strategy_returns) - 1
    benchmark_returns = np.cumprod(1 + returns) - 1

    # NaN 또는 Inf 값이 있는지 확인하고 처리
    cumulative_returns = np.nan_to_num(cumulative_returns, nan=0.0, posinf=0.0, neginf=0.0)
    benchmark_returns = np.nan_to_num(benchmark_returns, nan=0.0, posinf=0.0, neginf=0.0)

    print(f"Strategy Final Return: {cumulative_returns[-1] * 100:.2f}%")
    print(f"Benchmark Final Return: {benchmark_returns[-1] * 100:.2f}%")

def main():
    for file_name in os.listdir(PREPROCESSED_DATA_DIR):
        if file_name.endswith('.csv'):
            corp_code = os.path.splitext(file_name)[0]
            file_path = os.path.join(PREPROCESSED_DATA_DIR, file_name)

            logging.info(f"Processing {corp_code}...")

            train_data, val_data, test_data, feature_columns, scaler = load_and_preprocess_data(file_path)
            X_train, y_train = create_sequences(train_data, SEQUENCE_LENGTH, feature_columns)
            X_val, y_val = create_sequences(val_data, SEQUENCE_LENGTH, feature_columns)
            X_test, y_test = create_sequences(test_data, SEQUENCE_LENGTH, feature_columns)

            model, history = build_and_train_model(X_train, y_train, X_val, y_val)
            y_pred = evaluate_model(model, X_test, y_test)
            calculate_and_compare_returns(test_data, y_pred)

            corp_model_dir = os.path.join(MODEL_DIR, corp_code)
            os.makedirs(corp_model_dir, exist_ok=True)
            model_path = os.path.join(corp_model_dir, f"{corp_code}_lstm_model.h5")
            scaler_path = os.path.join(corp_model_dir, f"{corp_code}_scaler.pkl")

            model.save(model_path)
            joblib.dump(scaler, scaler_path)
            logging.info(f"Model and scaler saved for {corp_code}.")

if __name__ == "__main__":
    main()
